char*gModuleName="\030All Steam V1.03 Academic";

/*******************************************************************************
         AAAAA    lll     lll    SSSSS     tt
        AA   AA    ll      ll   SS   SS    tt
        AA   AA    ll      ll   SS       tttttt  eeeee   aaaaa  mmm mmm
        AAAAAAA    ll      ll    SSSSS     tt   ee   ee      aa mmmmmmm
        AA   AA    ll      ll        SS    tt   eeeeeee  aaaaaa mm m mm
        AA   AA    ll      ll   SS   SS    tt   ee      aa   aa mm   mm
        AA   AA   llll    llll   SSSSS     tttt  eeeee   aaaaaa mm   mm
        ---------------------------------------------------------------
              Steam Property Excel Add-In by Dudley J. Benton

  V1.00 (11/17/2016): first release
  V1.01 (07/21/2016): fix bug in Tsat(T<Pmin,year) Psat(T<Tmin,year)
  V1.02 (11/30/2018): add SI units (use -year)
  V1.03 (03/06/2019): reorganize functions and change base units SI
        (05/16/2019): add conditional compilation statements to
                      facilitate the creation of static libraries

*******************************************************************************/

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#define _USE_MATH_DEFINES
#include <float.h>
#include <math.h>

#ifdef _WINDLL
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "xlcall.h"

/*******************************************************************************
  FUNCTION DEFINITIONS AND ARGUMENTS - ALL FUNCTIONS MUST BE DEFINED HERE

  Each function is defined by a group of entries ending in a NULL strings.
  Each entry is a character, the variable name, and description. The
  character tells what type of variable it is according to this list:

  A boolean (short int 0 or 1) passed by value
  B double passed by value
  C null-terminaed string passed by pointer
  D Pascal (length predecated) string passed by pointer
  E double passed by pointer
  F null-terminaed string passed by pointer
  G Pascal (length predecated) string passed by pointer
  H unsigned short (16-bit) int (i.e., WORD) passed by value
  I signed short (16-bit) int passed by value
  J signed long (32-bit) int passed by value
  K floating point array structure?
  L boolean (short int 0 or 1) passed by pointer
  M signed short (16-bit) int passed by pointer
  N signed long (32-bit) int passed by pointer
  O array structure {WORD,WORD,double*} passed as pointer
  P XLOPER passed as pointer used for variable type worksheet values and arrays
  R XLOPER passed as pointer used for values, arrays, and range references

*******************************************************************************/

struct{char type;char*name,*desc;}Funs[]={

  {'R',"Psat"        ,"saturation pressure [psia:MPa]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Tsat"        ,"saturation temperature [F:K]"},
  {'B',"Ps"          ,"saturation pressure [psia:MPa]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Vf"          ,"specific volume of saturated liquid [ft/lbm:cm/g]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Vg"          ,"specific volume of saturated vapor [ft/lbm:cm/g]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Hf"          ,"enthalpy of saturated liquid [BTU/lbm:kJ/kg]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Hg"          ,"enthalpy of saturated vapor [BTU/lbm:kJ/kg]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Sf"          ,"entropy of saturated liquid [BTU/lbm/R:kJ/kg/K]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"Sg"          ,"entropy of saturated vapor [BTU/lbm/R:kJ/kg/K]"},
  {'B',"Ts"          ,"saturation temperature [F:K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"VofTP"       ,"specific volume [ft/lbm:cm/g]"},
  {'B',"T"           ,"temperature [F:K]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"HofTP"       ,"enthalpy [BTU/lbm:kJ/kg]"},
  {'B',"T"           ,"temperature [F:K]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"SofTP"       ,"entropy [BTU/lbm/R:kJ/kg/K]"},
  {'B',"T"           ,"temperature [F:K]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"TofPH"       ,"temperature [F:K]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'B',"H"           ,"enthalpy [BTU/lbm:kJ/kg]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"SofPH"       ,"entropy [BTU/lbm/R:kJ/kg/K]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'B',"H"           ,"enthalpy [BTU/lbm:kJ/kg]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"TofPS"       ,"temperature [F:K]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'B',"S"           ,"entropy [BTU/lbm/R:kJ/kg/K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  {'R',"HofPS"       ,"enthalpy [BTU/lbm:kJ/kg]"},
  {'B',"P"           ,"pressure [psia:MPa]"},
  {'B',"S"           ,"entropy [BTU/lbm/R:kJ/kg/K]"},
  {'J',"formulation" ,"67=ASME, 69=KKHM, 84=NBS/NRC, 95=SF95, 97=IF97"},
  { 0 ,NULL          ,NULL},

  { 0 ,NULL          ,NULL}};

/*******************************************************************************
  THIS SECTION IS ALWAYS THE SAME
*******************************************************************************/

int APIENTRY DllMain(HANDLE hModule,DWORD fdwReason,LPVOID lpReserved)
  {
  return(1);
  }

char**BuildFunctionParms()
  {
  char args[256],**funs,type[31];
  int i,j,m,n;

  for(m=j=i=0;Funs[i].type;i++,j++)
    {
    for(n=0;Funs[i+1+n].type;n++)
      ;
    i+=n+1;
    if(sizeof(void*)>4)
      m+=n+5;
    else
      m+=min(20,n)+5;
    }

  if((funs=calloc(m+1,sizeof(char*)))==NULL)
    return(NULL);

  for(m=i=0;Funs[i].type;i++)
    {
    funs[m++]=_strdup(Funs[i].name);
    memset(type,0,sizeof(type));
    type[0]=Funs[i].type;
    args[0]=0;
    for(n=0;Funs[i+1+n].type;n++)
      {
      type[n+1]=Funs[i+1+n].type;
      if(n)
        strcat(args,",");
      strcat(args,Funs[i+1+n].name);
      }
    if(n>30)
      return(NULL);
    funs[m++]=_strdup(type);
    funs[m++]=_strdup(args);
    funs[m++]=_strdup(Funs[i].desc);
    for(j=0;j<n;j++)
      {
      if(j==19&&n>20&&sizeof(void*)==4)
        {
        strcpy(args,Funs[i+1+j].desc);
        strcat(args,", etc.");
        funs[m++]=_strdup(args);
        }
      else if(j<20||sizeof(void*)>4)
        funs[m++]=_strdup(Funs[i+1+j].desc);
      }
    m++;
    i+=n+1;
    }

  return(funs);
  }

char**functionParms;

__declspec(dllexport) int xlAutoOpen(void)
  {
  int i,j,k,l,rv;
  XLOPER xConfirm,xDLL,*xStr,**xPtr,xStrT,xIntT;

  if((functionParms=BuildFunctionParms())==NULL)
    {
    xStrT.val.str=" Error: can't build function params!";
    xStrT.val.str[0]=(char)(strlen(xStrT.val.str)-1);
    Excel4(xlcAlert,0,2,&xStrT,&xIntT);
    return(0);
    }

  for(k=i=0;functionParms[i];i++)
    {
    for(j=0;functionParms[i];i++)
      j++;
    k=max(k,j);
    }
  k+=5;

  xStr=calloc(k,sizeof(XLOPER));
  xPtr=calloc(k+1,sizeof(XLOPER*));
  xIntT.xltype=xltypeInt;
  xIntT.val.w=2;
  xStrT.xltype=xltypeStr;
  Excel4(xlGetName,&xDLL,0);
  xPtr[0]=&xDLL;
  for(i=0;functionParms[i];i++)
    {
    for(j=0;functionParms[i];i++)
      {
      l=(int)strlen(functionParms[i]);
      xStr[j].xltype=xltypeStr;
      xStr[j].val.str=malloc(l+1);
      memcpy(xStr[j].val.str+1,functionParms[i],l);
      xStr[j].val.str[0]=l;
      xPtr[j+1]=xStr+j;
      j++;
      if(j==2)
        {
        l=(int)strlen(functionParms[i-1]);
        xStr[j].xltype=xltypeStr;
        xStr[j].val.str=malloc(l+1);
        memcpy(xStr[j].val.str+1,functionParms[i-1],l);
        xStr[j].val.str[0]=l;
        xPtr[j+1]=xStr+j;
        j++;
        }
      if(j==4)
        {
        l=1;
        xStr[j].xltype=xltypeStr;
        xStr[j].val.str=malloc(l+1);
        memcpy(xStr[j].val.str+1,"1",l);
        xStr[j].val.str[0]=l;
        xPtr[j+1]=xStr+j;
        j++;
        }
      if(j==5)
        {
        l=(int)strlen(gModuleName+1);
        xStr[j].xltype=xltypeStr;
        xStr[j].val.str=malloc(l+1);
        memcpy(xStr[j].val.str+1,gModuleName+1,l);
        xStr[j].val.str[0]=l;
        xPtr[j+1]=xStr+j;
        j++;
        }
      if(j==6)
        {
        l=0;
        xStr[j].xltype=xltypeStr;
        xStr[j].val.str=malloc(1);
        xStr[j].val.str[0]=0;
        xPtr[j+1]=xStr+j;
        j++;
        }
      if(j==7)
        {
        l=0;
        xStr[j].xltype=xltypeStr;
        xStr[j].val.str=malloc(1);
        xStr[j].val.str[0]=0;
        xPtr[j+1]=xStr+j;
        j++;
        }
      }
    rv=Excel4v(xlfRegister,&xConfirm,j+1,&xPtr[0]);
    for(k=0;k<j;k++)
      free(xStr[k].val.str);
    if(rv!=xlretSuccess)
      {
      xStrT.val.str=" Error: Failed to register the XLL!";
      xStrT.val.str[0]=(char)(strlen(xStrT.val.str)-1);
      Excel4(xlcAlert,0,2,&xStrT,&xIntT);
      return(0);
      }
    }
  Excel4(xlFree,0,1,&xDLL);
  return(1);
  }

__declspec(dllexport) int xlAutoClose(void)
  {
  return(1);
  }

__declspec(dllexport) LPXLOPER xlAutoRegister(LPXLOPER pxName)
  {
  static XLOPER xRegId;
  xRegId.xltype=xltypeErr;
  xRegId.val.err=xlerrValue;
  return(&xRegId);
  }

__declspec(dllexport) int xlAutoAdd(void)
  {
  static XLOPER xStrT,xIntT;
  char szBuf[255];
  sprintf(szBuf," %s Loaded!\nbuilt on %s at %s",gModuleName+1,__DATE__,__TIME__);
  szBuf[0]=(char)(strlen(szBuf)-1);
  xStrT.xltype=xltypeStr;
  xStrT.val.str=szBuf;
  xIntT.xltype=xltypeInt;
  xIntT.val.w=2;
  Excel4(xlcAlert,0,2,&xStrT,&xIntT);
  return(1);
  }

__declspec(dllexport) int xlAutoRemove(void)
  {
  char szBuf[255];
  static XLOPER xStr,xInt;
  sprintf(szBuf," %s Removed!\nbuilt on %s at %s",gModuleName+1,__DATE__,__TIME__);
  szBuf[0]=(char)(strlen(szBuf)-1);
  xStr.xltype=xltypeStr;
  xStr.val.str=szBuf;
  xInt.xltype=xltypeInt;
  xInt.val.w=2;
  Excel4(xlcAlert,0,2,&xStr,&xInt);
  return(1);
  }

void CopyPascalString(char*str1,char*str2)
  {
  short i,len;
  len=str1[0];
  for(i=0;i<=len;i++)
    str2[i]=str1[i];
  }

void ConcatenatePascalString(char*str1,char*str2)
  {
  short i,len1,len2;
  len1=str1[0];
  len2=str2[0];
  str1[0]=len1+len2;
  for(i=0;i<len2;i++)
    str1[len1+i]=str2[i+1];
  }

__declspec(dllexport) LPXLOPER xlAddInManagerInfo(LPXLOPER xAction)
  {
  static XLOPER xInfo,xIntAction,xInt;
  static char tempBuf[255];
  xInt.xltype=xltypeInt;
  xInt.val.w=xltypeInt;
  Excel4(xlCoerce,&xIntAction,2,xAction,&xInt);
  if(xIntAction.val.w==1)
   {
   xInfo.xltype=xltypeStr;
   xInfo.val.str=gModuleName;
   }
  else
   {
   xInfo.xltype=xltypeErr;
   xInfo.val.err=xlerrValue;
   }
  return((LPXLOPER) &xInfo);
  }

XLOPER*ReturnValue(double value)
  {
  static XLOPER result;
  result.xltype=xltypeNum;
  result.val.num=value;
  return(&result);
  }

#define XL0PER __declspec(dllexport)LPXLOPER

#else

#define ReturnValue(x) x
#define XL0PER double

#endif

/*******************************************************************************
  PUT THE EXCEL FUNCTIONS HERE
*******************************************************************************/

static double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

static double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

#include "KKHM.h"
#include "IF67.h"
#include "NBSNRC.h"
#include "SF95.h"
#include "IF97.h"

static double _Psat(double Ts,int year)
  {
  double Ps;
  if(abs(year)==67)
    Ps=Psat67(Ts);
  else if(abs(year)==69)
    Ps=Psat69(Ts);
  else if(abs(year)==84)
    Ps=Psat84(Ts);
  else if(abs(year)==95)
    Ps=Psat95(Ts);
  else if(abs(year)==97)
    Ps=Psat97(Ts);
  else
    Ps=-999.;
  return(Ps);
  }

XL0PER Psat(double Ts,int year)
  {
  double Ps;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  Ps=_Psat(Ts,year);
  return(ReturnValue(year>0?Ps*145.0377377:Ps));
  }

static double _Tsat(double Ps,int year)
  {
  double Ts;
  if(abs(year)==67)
    Ts=Tsat67(Ps);
  else if(abs(year)==69)
    Ts=Tsat69(Ps);
  else if(abs(year)==84)
    Ts=Tsat84(Ps);
  else if(abs(year)==95)
    Ts=Tsat95(Ps);
  else if(abs(year)==97)
    Ts=Tsat97(Ps);
  else
    Ts=-999.;
  return(Ts);
  }

XL0PER Tsat(double Ps,int year)
  {
  double Ts;
  if(year>0)
    Ps/=145.0377377;
  Ts=_Tsat(Ps,year);
  return(ReturnValue(year>0?Ts*1.8-459.67:Ts));
  }

enum{liquid,vapor,supercritical};

XL0PER Vf(double Ts,int year)
  {
  double Vf;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  if(abs(year)==67)
    Vf=V67(Ts,0.,liquid);
  else if(abs(year)==69)
    Vf=1./R69(Ts,0.,liquid);
  else if(abs(year)==84)
    Vf=1./R84(Ts,0.,liquid);
  else if(abs(year)==95)
    Vf=1./R95(Ts,0.,liquid);
  else if(abs(year)==97)
    Vf=V97(Ts,0.,liquid);
  else
    return(ReturnValue(-999.));
  return(ReturnValue(year>0?Vf/62.42796058:Vf));
  }

XL0PER Vg(double Ts,int year)
  {
  double Vg;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  if(abs(year)==67)
    Vg=V67(Ts,0.,vapor);
  else if(abs(year)==69)
    Vg=1./R69(Ts,0.,vapor);
  else if(abs(year)==84)
    Vg=1./R84(Ts,0.,vapor);
  else if(abs(year)==95)
    Vg=1./R95(Ts,0.,vapor);
  else if(abs(year)==97)
    Vg=V97(Ts,0.,vapor);
  else
    return(ReturnValue(-999.));
  return(ReturnValue(year>0?Vg/62.42796058:Vg));
  }

static double _Hf(double Ts,int year)
  {
  double Hf,Vf;
  if(abs(year)==67)
    Hf=H67(Ts,0.,liquid);
  else if(abs(year)==69)
    {
    Vf=1./R69(Ts,0.,liquid);
    Hf=H69(Ts,1./Vf);
    }
  else if(abs(year)==84)
    {
    Vf=1./R84(Ts,0.,liquid);
    Hf=H84(Ts,1./Vf);
    }
  else if(abs(year)==95)
    {
    Vf=1./R95(Ts,0.,liquid);
    Hf=H95(Ts,1./Vf);
    }
  else if(abs(year)==97)
    Hf=H97(Ts,0.,liquid);
  else
    Hf=-999.;
  return(Hf);
  }

XL0PER Hf(double Ts,int year)
  {
  double Hf;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  Hf=_Hf(Ts,year);
  return(ReturnValue(year>0?Hf/2.326:Hf));
  }

static double _Hg(double Ts,int year)
  {
  double Hg,Vg;
  if(abs(year)==67)
    Hg=H67(Ts,0.,vapor);
  else if(abs(year)==69)
    {
    Vg=1./R69(Ts,0.,vapor);
    Hg=H69(Ts,1./Vg);
    }
  else if(abs(year)==84)
    {
    Vg=1./R84(Ts,0.,vapor);
    Hg=H84(Ts,1./Vg);
    }
  else if(abs(year)==95)
    {
    Vg=1./R95(Ts,0.,vapor);
    Hg=H95(Ts,1./Vg);
    }
  else if(abs(year)==97)
    Hg=H97(Ts,0.,vapor);
  else
    Hg=-999.;
  return(Hg);
  }

XL0PER Hg(double Ts,int year)
  {
  double Hg;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  Hg=_Hg(Ts,year);
  return(ReturnValue(year>0?Hg/2.326:Hg));
  }

static double _Sf(double Ts,int year)
  {
  double Sf,Vf;
  if(abs(year)==67)
    Sf=S67(Ts,0.,liquid);
  else if(abs(year)==69)
    {
    Vf=1./R69(Ts,0.,liquid);
    Sf=S69(Ts,1./Vf);
    }
  else if(abs(year)==84)
    {
    Vf=1./R84(Ts,0.,liquid);
    Sf=S84(Ts,1./Vf);
    }
  else if(abs(year)==95)
    {
    Vf=1./R95(Ts,0.,liquid);
    Sf=S95(Ts,1./Vf);
    }
  else if(abs(year)==97)
    Sf=S97(Ts,0.,liquid);
  else
    Sf=-999.;
  return(Sf);
  }

XL0PER Sf(double Ts,int year)
  {
  double Sf;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  Sf=_Sf(Ts,year);
  return(ReturnValue(year>0?Sf/4.186:Sf));
  }

static double _Sg(double Ts,int year)
  {
  double Sg,Vg;
  if(abs(year)==67)
    Sg=S67(Ts,0.,vapor);
  else if(abs(year)==69)
    {
    Vg=1./R69(Ts,0.,vapor);
    Sg=S69(Ts,1./Vg);
    }
  else if(abs(year)==84)
    {
    Vg=1./R84(Ts,0.,vapor);
    Sg=S84(Ts,1./Vg);
    }
  else if(abs(year)==95)
    {
    Vg=1./R95(Ts,0.,vapor);
    Sg=S95(Ts,1./Vg);
    }
  else if(abs(year)==97)
    Sg=S97(Ts,0.,vapor);
  else
    Sg=-999.;
  return(Sg);
  }

XL0PER Sg(double Ts,int year)
  {
  double Sg;
  if(year>0)
    Ts=(Ts+459.67)/1.8;
  Sg=_Sg(Ts,year);
  return(ReturnValue(year>0?Sg/4.186:Sg));
  }

XL0PER VofTP(double T,double P,int year)
  {
  double V;
  if(year>0)
    {
    T=(T+459.67)/1.8;
    P/=145.0377377;
    }
  if(abs(year)==67)
    V=V67(T,P,supercritical);
  else if(abs(year)==69)
    V=1./R69(T,P,supercritical);
  else if(abs(year)==84)
    V=1./R84(T,P,supercritical);
  else if(abs(year)==95)
    V=1./R95(T,P,supercritical);
  else if(abs(year)==97)
    V=V97(T,P,supercritical);
  else
    return(ReturnValue(-999.));
  return(ReturnValue(year>0?V/62.42796058:V));
  }

static double _HofTP(double T,double P,int year)
  {
  double H,V;
  if(abs(year)==67)
    H=H67(T,P,supercritical);
  else if(abs(year)==69)
    {
    V=1./R69(T,P,supercritical);
    H=H69(T,1./V);
    }
  else if(abs(year)==84)
    {
    V=1./R84(T,P,supercritical);
    H=H84(T,1./V);
    }
  else if(abs(year)==95)
    {
    V=1./R95(T,P,supercritical);
    H=H95(T,1./V);
    }
  else if(abs(year)==97)
    H=H97(T,P,supercritical);
  else
    H=-999.;
  return(H);
  }

XL0PER HofTP(double T,double P,int year)
  {
  double H;
  if(year>0)
    {
    T=(T+459.67)/1.8;
    P/=145.0377377;
    }
  H=_HofTP(T,P,year);
  return(ReturnValue(year>0?H/2.326:H));
  }

static double _SofTP(double T,double P,int year)
  {
  double S,V;
  if(abs(year)==67)
    S=S67(T,P,supercritical);
  else if(abs(year)==69)
    {
    V=1./R69(T,P,supercritical);
    S=S69(T,1./V);
    }
  else if(abs(year)==84)
    {
    V=1./R84(T,P,supercritical);
    S=S84(T,1./V);
    }
  else if(abs(year)==95)
    {
    V=1./R95(T,P,supercritical);
    S=S95(T,1./V);
    }
  else if(abs(year)==97)
    S=S67(T,P,supercritical);
  else
    S=-999.;
  return(S);
  }

XL0PER SofTP(double T,double P,int year)
  {
  double S;
  if(year>0)
    {
    T=(T+459.67)/1.8;
    P/=145.0377377;
    }
  S=_SofTP(T,P,year);
  return(ReturnValue(year>0?S/4.1868:S));
  }

static double _TofPH(double P,double H,int year)
  {
  int i;
  double H1,H2,H3,Hf,Hg,Pc,T1,T2,T3,Ts;
  Pc=_Psat(999.,year);
  if(P<Pc)
    {
    Ts=_Tsat(P,year);
    Hf=_Hf(Ts,year);
    Hg=_Hg(Ts,year);
    if(H<Hf)
      {
      T1=273.15;
      T2=Ts;
      }
    else if(H>Hg)
      {
      T1=Ts;
      T2=1500.;
      }
    else
      return(Ts);
    }
  T1=273.15;
  T2=1500.;
  H1=_HofTP(T1,P,year);
  H2=_HofTP(T2,P,year);
  for(i=0;i<32;i++)
    {
    if(fabs(H2-H1)<0.001)
      break;
    T3=(T1*(H2-H)-T2*(H1-H))/(H2-H1);
    H3=_HofTP(T3,P,year);
    if(fabs(H3-H)<0.01)
      break;
    if(fabs(H1-H)<fabs(H2-H))
      {
      T2=T3;
      H2=H3;
      }
    else
      {
      T1=T3;
      H1=H3;
      }
    }
  return(T3);
  }

XL0PER TofPH(double P,double H,int year)
  {
  double T;
  if(year>0)
    {
    P/=145.0377377;
    H*=2.326;
    }
  T=_TofPH(P,H,year);
  return(ReturnValue(year>0?T*1.8-459.67:T));
  }

static double _SofPH(double P,double H,int year)
  {
  double Hf,Hg,Pc,Sf,Sg,T,Ts,X;
  Pc=_Psat(999.,year);
  if(P<Pc)
    {
    Ts=_Tsat(P,year);
    Hf=_Hf(Ts,year);
    Hg=_Hg(Ts,year);
    if(H<Hf)
      {
      T=_TofPH(P,H,year);
      return(_SofTP(T,P,year));
      }
    else if(H>Hg)
      {
      T=_TofPH(P,H,year);
      return(_SofTP(T,P,year));
      }
    else
      {
      Sf=_Sf(Ts,year);
      Sg=_Sg(Ts,year);
      X=(H-Hf)/(Hg-Hf);
      return(Sf+X*(Sg-Sf));
      }
    }
  T=_TofPH(P,H,year);
  return(_SofTP(T,P,year));
  }

XL0PER SofPH(double P,double H,int year)
  {
  double S;
  if(year>0)
    {
    P/=145.0377377;
    H*=2.326;
    }
  S=_SofPH(P,H,year);
  return(ReturnValue(year>0?S/4.1868:S));
  }

static double _TofPS(double P,double S,int year)
  {
  int i;
  double Pc,S1,S2,S3,Sf,Sg,T1,T2,T3,Tm,Ts,Tx;
  Pc=_Psat(999.,year);
  if(P<Pc)
    {
    Ts=_Tsat(P,year);
    Sf=_Sf(Ts,year);
    Sg=_Sg(Ts,year);
    if(S<Sf)
      {
      Tm=273.15;
      Tx=Ts;
      }
    else if(S>Sg)
      {
      Tm=Ts;
      Tx=1500.;
      }
    else
      return(Ts);
    }
  else
    {
    Tm=273.15;
    Tx=1500.;
    }
  T1=Tm;
  T2=Tx;
  T3=(T1+T2)/2.;
  S1=_SofTP(T1,P,year);
  S2=_SofTP(T2,P,year);
  for(i=0;i<32;i++)
    {
    if(fabs(S2-S1)<0.000001)
      break;
    T3=fmax((Tm+T3)/2.,fmin((Tx+T3)/2.,(T1*(S2-S)-T2*(S1-S))/(S2-S1)));
    S3=_SofTP(T3,P,year);
    if(fabs(S3-S)<0.00001)
      break;
    if(fabs(S1-S)<fabs(S2-S))
      {
      T2=T3;
      S2=S3;
      }
    else
      {
      T1=T3;
      S1=S3;
      }
    }
  return(T3);
  }

XL0PER TofPS(double P,double S,int year)
  {
  double T;
  if(year>0)
    {
    P/=145.0377377;
    S*=4.1868;
    }
  T=_TofPS(P,S,year);
  return(ReturnValue(year>0?T*1.8-459.67:T));
  }

static double _HofPS(double P,double S,int year)
  {
  double Hf,Hg,Pc,Sf,Sg,T,Ts,X;
  Pc=_Psat(999.,year);
  if(P<Pc)
    {
    Ts=_Tsat(P,year);
    Sf=_Sf(Ts,year);
    Sg=_Sg(Ts,year);
    if(S<Sf)
      {
      T=_TofPS(P,S,year);
      return(_HofTP(T,P,year));
      }
    else if(S>Sg)
      {
      T=_TofPS(P,S,year);
      return(_HofTP(T,P,year));
      }
    else
      {
      Hf=_Hf(Ts,year);
      Hg=_Hg(Ts,year);
      X=(S-Sf)/(Sg-Sf);
      return(Hf+X*(Hg-Hf));
      }
    }
  T=_TofPS(P,S,year);
  return(_HofTP(T,P,year));
  }

XL0PER HofPS(double P,double S,int year)
  {
  double H;
  if(year>0)
    {
    P/=145.0377377;
    S*=4.1868;
    }
  H=_HofPS(P,S,year);
  return(ReturnValue(year>0?H/2.326:H));
  }
