/* simple single-pressure HRSG
   using SI units
   developed by Dudley J. Benton */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "..\\AllSteam\\AllSteam.h"

int year=-97; /* steam properties (must be 67, 69, 84, 95, or 97) */
              /* use +year for English and -year for SI units */

/* GT exhaust mole fractions, molecular weights, and specific heats */

struct{double N2,O2,CO2,H2O,Ar;}mf={0.746,0.124,0.038,0.083,0.009};
struct{double N2,O2,CO2,H2O,Ar;}MW={28.00,32.00,44.00,18.00,40.00};
struct{double N2,O2,CO2,H2O,Ar;}CP={1.040,0.919,0.844,1.930,0.520};

/* GT exhaust temperature and flow */

struct{double F,T;}GT={4000.,600.};

/* feedwater inlet temperature */

double Tfw=20.;

/* steam outlet pressure (controlled by steam turbine swallowing capacity) */

double Pstm=8000.;

/* pressure drops */

struct{double sh,ec;}dP={250.,750.};

/* desired temperatures (superheat and subcooling) */

struct{double sh,ec;}dT={250.,5.};

/* overall UA */

double UAtarget=8000.;

double LMTD1(double dT1,double dT2)
  {
  if(dT1<=0.||dT2<=0.)
    return(0.001);
  if(fabs(dT1-dT2)<0.001)
    return(sqrt(dT1*dT2));
  return((dT1-dT2)/log(dT1/dT2));
  }

void Solver1()
  {
  int iter;
  double Cp,dTec,dTev,dTsh,Qec,Qev,Qsh;
  double F1,F2,Fstm,UAec,UAev,UAsh,UAtotal;
  double H1,H2,H3,H4,P1,P2,P3,P4;
  double T1,T2,T3,T4,T5,T6,T7,T8;
  Cp=(mf.N2*MW.N2*CP.N2+mf.O2*MW.O2*CP.O2+mf.CO2*MW.CO2*CP.CO2+mf.H2O*MW.H2O*CP.H2O+mf.Ar*MW.Ar*CP.Ar)
    /(mf.N2*MW.N2+mf.O2*MW.O2+mf.CO2*MW.CO2+mf.H2O*MW.H2O+mf.Ar*MW.Ar);
  P4=Pstm;
  P3=P4+dP.sh;
  P2=P3;
  P1=P2+dP.ec;
  T1=Tfw;
  T3=Tsat(P3/1000.,year)-273.15;
  T2=T3-dT.ec;
  T4=T3+dT.sh;
  T5=GT.T;
  H1=HofTP(T1+273.15,P1/1000.,year);
  H2=HofTP(T2+273.15,P2/1000.,year);
  H3=Hg(T3+273.15,year);
  H4=HofTP(T4+273.15,P4/1000.,year);
  F1=1.;
  F2=GT.F/8.;
  printf("it flo  UA\n");
  for(iter=0;iter<32;iter++)
    {
    Fstm=(F1+F2)/2.;
    Qec=Fstm*(H2-H1);
    Qev=Fstm*(H3-H2);
    Qsh=Fstm*(H4-H3);
    T6=T5-Qsh/GT.F/Cp;
    T7=T6-Qev/GT.F/Cp;
    T8=T7-Qec/GT.F/Cp;
    dTsh=LMTD1(T5-T4,T6-T3);
    dTev=LMTD1(T6-T3,T7-T2);
    dTec=LMTD1(T7-T2,T8-T1);
    UAsh=Qsh/dTsh;
    UAev=Qev/dTev;
    UAec=Qec/dTec;
    UAtotal=UAsh+UAev+UAec;
    printf("%2i %3.0lf %4.0lf\n",iter,Fstm,UAtotal);
    if(UAtotal<UAtarget)
      F1=Fstm;
    else
      F2=Fstm;
    }
  }

double LMTD2(double dT1,double dT2)
  {
  if(dT1<=0.||dT2<=0.)
    return(0.);
  if(fabs(dT1-dT2)<0.001)
    return(sqrt(dT1*dT2));
  return((dT1-dT2)/log(dT1/dT2));
  }

void Solver2()
  {
  int iter;
  double Cp,dTec,dTev,dTsh,Qec,Qev,Qsh;
  double Fstm,UAec,UAev,UAsh,UAtotal;
  double H1,H2,H3,H4,P1,P2,P3,P4;
  double T1,T2,T3,T4,T5,T6,T7,T8;
  Cp=(mf.N2*MW.N2*CP.N2+mf.O2*MW.O2*CP.O2+mf.CO2*MW.CO2*CP.CO2+mf.H2O*MW.H2O*CP.H2O+mf.Ar*MW.Ar*CP.Ar)
    /(mf.N2*MW.N2+mf.O2*MW.O2+mf.CO2*MW.CO2+mf.H2O*MW.H2O+mf.Ar*MW.Ar);
  P4=Pstm;
  P3=P4+dP.sh;
  P2=P3;
  P1=P2+dP.ec;
  T1=Tfw;
  T3=Tsat(P3/1000.,year)-273.15;
  T2=T3-dT.ec;
  T4=T3+dT.sh;
  T5=GT.T;
  H1=HofTP(T1+273.15,P1/1000.,year);
  H2=HofTP(T2+273.15,P2/1000.,year);
  H3=Hg(T3+273.15,year);
  H4=HofTP(T4+273.15,P4/1000.,year);
  Fstm=50.;
  printf("it flo  UA\n");
  for(iter=0;iter<32;iter++)
    {
    Qec=Fstm*(H2-H1);
    Qev=Fstm*(H3-H2);
    Qsh=Fstm*(H4-H3);
    if(Qec*Qev*Qsh<=0.)
      {
      Fstm/=2.;
      continue;
      }
    T6=T5-Qsh/GT.F/Cp;
    T7=T6-Qev/GT.F/Cp;
    T8=T7-Qec/GT.F/Cp;
    dTsh=LMTD2(T5-T4,T6-T3);
    dTev=LMTD2(T6-T3,T7-T2);
    dTec=LMTD2(T7-T2,T8-T1);
    if(dTsh*dTev*dTec<=0.)
      {
      Fstm/=2.;
      continue;
      }
    UAsh=Qsh/dTsh;
    UAev=Qev/dTev;
    UAec=Qec/dTec;
    UAtotal=UAsh+UAev+UAec;
    printf("%2i %3.0lf %4.0lf\n",iter,Fstm,UAtotal);
    if(fabs(UAtotal/UAtarget-1.)<0.0001)
      break;
    Fstm*=(1.+UAtarget/UAtotal)/2.;
    }
  }

int main(int argc,char**argv,char**envp)
  {
//Solver1();
  Solver2();
  return(0);
  }
