/* simple two-pressure HRSG
   using English units
   developed by Dudley J. Benton */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "..\\AllSteam\\AllSteam.h"

int year=97; /* steam properties (must be 67, 69, 84, 95, or 97) */
             /* use +year for English and -year for SI units */

/* GT exhaust mole fractions, molecular weights, and specific heats */

struct{double N2,O2,CO2,H2O,Ar;}mf={0.746,0.124,0.038,0.083,0.009};
struct{double N2,O2,CO2,H2O,Ar;}MW={28.00,32.00,44.00,18.00,40.00};
struct{double N2,O2,CO2,H2O,Ar;}CP={0.249,0.219,0.193,0.443,0.124};

/* GT exhaust temperature and flow */

struct{double F,T;}GT={1000000.,1050.};

/* feedwater inlet temperature */

double Tfw=60.;

/* steam outlet pressure (controlled by steam turbine swallowing capacity) */

struct{double HP,LP;}Pstm={1800.,75.};

/* pressure drops */

struct{double sh1,sh2,ec1,ec2,ec3;}dP={5.,25.,100.,50.,50.};

/* desired temperature differences (superheat and subcooling) */

struct{double sh1,sh2,ec1,ec2,ec3;}dT={250.,250.,5.,150.,20.};

/* overall UA */

struct{double LP,HP;}UAtarget={2E5,6E5};

double LMTD(double dT1,double dT2)
  {
  if(dT1<=0.||dT2<=0.)
    return(0.);
  if(fabs(dT1-dT2)<0.001)
    return(sqrt(dT1*dT2));
  return((dT1-dT2)/log(dT1/dT2));
  }

void Solver()
  {
  int iter;
  double Cp,dTsh2,dTev2,dTec3,dTec2,dTsh1,dTev1,dTec1,Fhp,Flp,Hbfw,Hfw,Hec1,
    Hev1,Hsh1,Hec2,Hec3,Hev2,Hsh2,Pec1,Pec2,Pec3,Pev1,Pev2,Psh1,Psh2,Qec1,Qec2,
    Qec3,Qev1,Qev2,Qsh1,Qsh2,T1,T2,T3,T4,T5,T6,T7,T8,Tbfw,Tec1,Tec2,Tec3,Tev1,
    Tev2,Tsh1,Tsh2,UAhp,UAlp;
  Cp=(mf.N2*MW.N2*CP.N2+mf.O2*MW.O2*CP.O2+mf.CO2*MW.CO2*CP.CO2+mf.H2O*MW.H2O*CP.H2O+mf.Ar*MW.Ar*CP.Ar)
    /(mf.N2*MW.N2+mf.O2*MW.O2+mf.CO2*MW.CO2+mf.H2O*MW.H2O+mf.Ar*MW.Ar);
  Psh1=Pstm.LP;
  Pev1=Psh1+dP.sh1;
  Pec1=Pev1+dP.ec1;
  Psh2=Pstm.HP;
  Pev2=Psh2+dP.sh2;
  Pec3=Pev2+dP.ec3;
  Pec2=Pec3+dP.ec2;
  Tev1=Tsat(Pev1,year);
  Tec1=Tev1-dT.ec1;
  Tsh1=Tev1+dT.sh1;
  Tev2=Tsat(Pev2,year);
  Tec3=Tev2-dT.ec3;
  Tec2=Tev2-dT.ec2;
  Tsh2=Tev2+dT.sh2;
  Hfw =HofTP(Tfw ,Pec1,year);
  Hec1=HofTP(Tec1,Pev1,year);
  Hev1=Hg(Tev1,year);
  Hsh1=HofTP(Tsh1,Psh1,year);
  Tbfw=Tev1;
  Hbfw=Hf(Tbfw,year);
  Hec2=HofTP(Tec2,Pec2,year);
  Hec3=HofTP(Tec3,Pec3,year);
  Hev2=Hg(Tev2,year);
  Hsh2=HofTP(Tsh2,Psh2,year);
  printf("it  Flp    Fhp   UAlp   UAhp\n");
  Flp=GT.F/100.;
  Fhp=GT.F/ 20.;
  for(iter=0;iter<32;iter++)
    {
    Qec1=(Flp+Fhp)*(Hec1-Hfw );
    Qev1= Flp     *(Hev1-Hec1);
    Qsh1= Flp     *(Hsh1-Hev1);
    Qec2= Fhp     *(Hec2-Hbfw);
    Qec3= Fhp     *(Hec3-Hec2);
    Qev2= Fhp     *(Hev2-Hec3);
    Qsh2= Fhp     *(Hsh2-Hev2);
    if(Qec1<=0.||Qev1<=0.||Qsh1<=0.||Qec2<=0.||Qec3<=0.||Qev2<=0.||Qsh2<=0.)
      {
      Flp/=2.;
      Fhp/=2.;
      continue;
      }
    T1=GT.T;
    T2=T1-Qsh2/GT.F/Cp;
    T3=T2-Qev2/GT.F/Cp;
    T4=T3-Qec3/GT.F/Cp;
    T5=T4-Qsh1/GT.F/Cp;
    T6=T5-Qec2/GT.F/Cp;
    T7=T6-Qev1/GT.F/Cp;
    T8=T7-Qec1/GT.F/Cp;
    dTsh2=LMTD(T1-Tsh2,T2-Tev2);
    dTev2=LMTD(T2-Tev2,T3-Tec3);
    dTec3=LMTD(T3-Tec3,T4-Tec2);
    dTsh1=LMTD(T4-Tsh1,T5-Tev1);
    dTec2=LMTD(T5-Tec2,T6-Tbfw);
    dTev1=LMTD(T6-Tev1,T7-Tec1);
    dTec1=LMTD(T7-Tec1,T8-Tfw );
    if(dTsh2<=0.||dTev2<=0.||dTec3<=0.||dTec2<=0.||dTsh1<=0.||dTev1<=0.||dTec1<=0.)
      {
      Flp/=2.;
      Fhp/=2.;
      continue;
      }
    UAhp=Qsh2/dTsh2+Qev2/dTev2+Qec3/dTec3+Qec2/dTec2;
    UAlp=Qsh1/dTsh1+Qev1/dTev1+Qec1/dTec1;
    printf("%2i %5.0lf %6.0lf %6.0lf %6.0lf\n",iter,Flp,Fhp,UAlp,UAhp);
    if(fabs(UAlp/UAtarget.LP-1.)<0.0001)
      if(fabs(UAhp/UAtarget.HP-1.)<0.0001)
        break;
    Flp*=(1.+UAtarget.LP/UAlp)/2.;
    Fhp*=(1.+UAtarget.HP/UAhp)/2.;
    }
  }

int main(int argc,char**argv,char**envp)
  {
  Solver();
  return(0);
  }
